/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 2 juillet 2007
   derniere modification le 08/10/2007 */


#include <SCE/SCEMinimal.h>

#include <SCE/core/SCECTexture.h>

#include <SCE/core/SCECFramebuffer.h>


/* ajoute le 05/10/2007 */
static SCE_CFramebuffer *binded = NULL;


/* ajoute le 05/10/2007 */
void SCE_CBindFramebuffer (SCE_CFramebuffer *fb)
{
    binded = fb;
    if (binded)
        glBindFramebufferEXT (GL_FRAMEBUFFER_EXT, binded->id);
    else
        glBindFramebufferEXT (GL_FRAMEBUFFER_EXT, 0);
}


/* ajoute le 05/10/2007 */
void SCE_CInitFramebuffer (SCE_CFramebuffer *fb)
{
    unsigned int i;

    fb->id = 0;
    fb->mask = 0;
    fb->x = fb->y = 0;
    fb->w = fb->h = 1;
    for (i=0; i<SCE_NUM_RENDER_BUFFERS; i++)
    {
        fb->buffers[i].id = 0;
        fb->buffers[i].tex = NULL;
        fb->buffers[i].user = SCE_TRUE;
    }
}

/* ajoute le 02/07/2007 */
/* revise le 05/10/2007 */
SCE_CFramebuffer* SCE_CCreateFramebuffer (void)
{
    SCE_CFramebuffer *fb = NULL;

    SCE_btstart ();
    fb = SCE_malloc (sizeof *fb);
    if (fb == NULL)
    {
        Logger_LogSrc ();
        SCE_btend ();
        return NULL;
    }

    SCE_CInitFramebuffer (fb);
    glGenFramebuffersEXT (1, &fb->id);

    SCE_btend ();
    return fb;
}

/* ... */
#define SCE_CDEFAULTFUNC(action)\
SCE_CFramebuffer *back = binded;\
SCE_CBindFramebuffer (fb);\
action;\
SCE_CBindFramebuffer (back);

#define SCE_CDEFAULTFUNCR(t, action)\
t r;\
SCE_CFramebuffer *back = binded;\
SCE_CBindFramebuffer (fb);\
r = action;\
SCE_CBindFramebuffer (back);\
return r;

/* ajoute le 02/07/2007 */
/* revise le 05/10/2007 */
void SCE_CDeleteFramebuffer (SCE_CFramebuffer *fb)
{
    SCE_CDEFAULTFUNC (SCE_CDeleteFramebuffer_ ())
}
void SCE_CDeleteFramebuffer_ (void)
{
    SCE_btstart ();
    if (binded)
    {
        unsigned int i;
        for(i=0; i<SCE_NUM_RENDER_BUFFERS; i++)
        {
            if (!binded->buffers[i].user)
                SCE_CDeleteTexture (binded->buffers[i].tex);
            
            if (binded->buffers[i].id)
                glDeleteRenderbuffersEXT (1, &binded->buffers[i].id);
        }
        
        glDeleteFramebuffersEXT (1, &binded->id);
        SCE_free (binded);
        binded = NULL;
    }
    SCE_btend ();
}


/* converti un simple ID de type d'attachement en type opengl */
static int SCE_CIDToGLBuffer (SCEuint type)
{
    if(type == SCE_COLOR_BUFFER)
        return GL_COLOR_ATTACHMENT0_EXT;
    else if(type == SCE_DEPTH_BUFFER)
        return GL_DEPTH_ATTACHMENT_EXT;
    else
        return GL_STENCIL_ATTACHMENT_EXT;
}


/* ajoute le 05/10/2007 */
void SCE_CAddRenderTexture (SCE_CFramebuffer *fb, SCEuint id, SCEenum target,
                            SCE_CTexture *tex, int mipmap)
{
    SCE_CDEFAULTFUNC (SCE_CAddRenderTexture_ (id, target, tex, mipmap))
}
/* ajoute le 02/07/2007 */
/* revise le 06/10/2007 */
void SCE_CAddRenderTexture_ (SCEuint id, SCEenum target,
                             SCE_CTexture *tex, int mipmap)
{
    /*SCE_CTexData *d = NULL;*/
    int type;

    SCE_btstart ();
    type = SCE_CIDToGLBuffer (id);

    /* si la precedente texture n'a pas ete envoyee par l'utilisateur,
       on la supprime */
    if (!binded->buffers[id].user)
        SCE_CDeleteTexture (binded->buffers[id].tex);

    /* si un buffer existe deja, on le supprime */
    if (binded->buffers[id].id)
    {
        glDeleteRenderbuffersEXT (1, &binded->buffers[id].id);
        binded->buffers[id].id = 0;
    }

    binded->buffers[id].tex = tex;

    /* verification du target */
    if (target < SCE_TEX_POSX || target > SCE_TEX_NEGZ)
        target = SCE_CGetTextureTarget (tex);
    else if (tex->target != SCE_TEX_CUBE)
    {
        /* target designant une face de cubemap, mais la texture
           n'est pas de type cubemap : erreur */
        Logger_Log (SCE_INVALID_ARG);
        Logger_LogMsg ("invalid target, the texture is not a cubemap");
        SCE_btend ();
        return;
    }

    /* et on lui assigne notre zolie render texture */
    glFramebufferTexture2DEXT (GL_FRAMEBUFFER_EXT, type,
                               target, tex->id, mipmap);

    /* si aucune color render texture n'existe, on desactive le tampon
       (me demandez pas pourquoi, j'en sais rien, allez faire un tour ici :
       http://www.cs.cmu.edu/afs/cs/academic/class/15462/web.06s/
       asst/project3/shadowmap/fbo.html
    */
    if (id == SCE_DEPTH_BUFFER && !binded->buffers[SCE_COLOR_BUFFER].tex)
        glDrawBuffer (GL_NONE);

    /* si en revanche un cree une color render texture, et qu'une depth
       render texture est deja creee, on reactive le tampon (logique la) */
    /*
    if(id == SCE_COLOR_BUFFER)
    {
        glDrawBuffer(GL_BACK);
    }*/
    /* TODO: gros fuck.... */

    /*
      status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
      if(status == GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT)
          printf("SCEngine: bouh!!!\n");
    */

    /* recuperation des dimensions depuis la texture */
    /* TODO: gne gne gne ?? pour le redimensionnement de tex ? */

    SCE_CClearRenderbuffer_ (id, 1);
    SCE_btend ();
}


/* ajoute le 06/10/2007 */
void SCE_CAddRenderBuffer (SCE_CFramebuffer *fb, SCEuint id,
                           int fmt, int w, int h)
{
    SCE_CDEFAULTFUNC (SCE_CAddRenderBuffer_ (id, fmt, w, h))
}
/* ajoute le 02/07/2007 */
/* revise le 06/10/2007 */
void SCE_CAddRenderBuffer_ (SCEuint id, int fmt, int w, int h)
{
    int type;

    SCE_btstart ();
    type = SCE_CIDToGLBuffer (id);

    /* assignation des valeurs par defaut */
    if (fmt <= 0)
    {
        if (id == SCE_DEPTH_BUFFER)
            fmt = GL_DEPTH_COMPONENT24;
        else if (id == SCE_STENCIL_BUFFER)
            fmt = GL_STENCIL_INDEX8_EXT;
    }

    /* si une texture existait deja, on la supprime */
    if (!binded->buffers[id].user)
    {
        SCE_CDeleteTexture (binded->buffers[id].tex);
        binded->buffers[id].tex = NULL;  /* ! important ! */
    }

    /* si le buffer n'existe pas, il faut le creer */
    if (!glIsRenderbufferEXT (binded->buffers[id].id))
        glGenRenderbuffersEXT (1, &binded->buffers[id].id);

    /* creation du render buffer */
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, binded->buffers[id].id);
    glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, fmt, w, h);
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, 0);

    /* on l'ajoute au FBO */
    glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, type, GL_RENDERBUFFER_EXT,
                                 binded->buffers[id].id);

    /* ici on force les nouvelles dimensions, pas cool... */
    binded->w = w;
    binded->h = h;

    SCE_CClearRenderbuffer_ (id, 1);
    SCE_btend ();
}


/* ajoute le 06/10/2007 */
int SCE_CCreateRenderTexture (SCE_CFramebuffer *fb, SCEuint id,
                              int pxf, int fmt, int type, int w, int h)
{
    SCE_CDEFAULTFUNCR (int,SCE_CCreateRenderTexture_ (id, pxf, fmt, type, w, h))
}
/* ajoute le 02/07/2007 */
/* revise le 03/07/2007 */
int SCE_CCreateRenderTexture_ (SCEuint id, int pxf, int fmt,
                               int type, int w, int h)
{
    SCE_CTexture *tex = NULL;
    SCE_CTexData data;

    SCE_btstart ();
    /* creation de la texture */
    tex = SCE_CCreateTexture (SCE_TEX_2D);
    if (!tex)
    {
        Logger_LogSrc ();
        SCE_btend ();
        return SCE_ERROR;
    }

    SCE_CInitTexData (&data);

    /* utilisation des valeurs par defaut en cas de non specification */
    if (pxf <= 0)
    {
        if (id == SCE_DEPTH_BUFFER)
            pxf = GL_DEPTH_COMPONENT24;
        else
            pxf = GL_RGBA;
    }
    if (fmt <= 0)
    {
        if (id == SCE_DEPTH_BUFFER)
            fmt = GL_DEPTH_COMPONENT;
        else
            fmt = GL_RGBA;
    }
    if (type <= 0)
        type = SCE_UNSIGNED_BYTE;

    /* assignation des donnees */
    data.w = w;
    data.h = h;
    data.type = type;
    data.pxf = pxf;
    data.fmt = fmt;

    SCE_CBindTexture (tex);
    if (SCE_CAddTextureTexDataDup_ (0, &data) < 0)
    {
        Logger_LogSrc ();
        SCE_CDeleteTexture (tex);
        SCE_btend ();
        return SCE_ERROR;
    }
    if (SCE_CBuildTexture_ (0, 0) < 0)
    {
        Logger_LogSrc ();
        SCE_CDeleteTexture (tex);
        SCE_btend ();
        return SCE_ERROR;
    }
    SCE_CBindTexture (NULL);

    /* ajout de la texture en tant que render buffer */
    SCE_CAddRenderTexture_ (id, SCE_TEX_2D, tex, 0);

    /* attention: la texture n'appartient pas a l'utilisateur */
    binded->buffers[id].user = SCE_FALSE;

    SCE_btend ();
    return SCE_OK;
}

/* ajoute le 03/07/2007 */
SCE_CTexture* SCE_CGetRenderTexture (SCE_CFramebuffer *fb, SCEuint id)
{
    return fb->buffers[id].tex;
}
/* ajoute le 06/10/2007 */
SCE_CTexture* SCE_CGetRenderTexture_ (SCEuint id)
{
    return binded->buffers[id].tex;
}


/* ajoute le 02/07/2007 */
/* revise le 06/10/2007 */
void SCE_CClearRenderbuffer (SCE_CFramebuffer *fb, SCEuint type, int active)
{
    /* TODO: a revoir le -= ... */
    switch (type)
    {
        case SCE_COLOR_BUFFER:
            if(active)
                fb->mask |= GL_COLOR_BUFFER_BIT;
            else
                fb->mask -= GL_COLOR_BUFFER_BIT;
            break;

        case SCE_DEPTH_BUFFER:
            if(active)
                fb->mask |= GL_DEPTH_BUFFER_BIT;
            else
                fb->mask -= GL_DEPTH_BUFFER_BIT;
            break;

        case SCE_STENCIL_BUFFER:
            if(active)
                fb->mask |= GL_STENCIL_BUFFER_BIT;
            else
                fb->mask -= GL_STENCIL_BUFFER_BIT;
    }
}
/* ajoute le 06/10/2007 */
void SCE_CClearRenderbuffer_ (SCEuint type, int active)
{
    SCE_CClearRenderbuffer (binded, type, active);
}


/* ajoute le 02/07/2007 */
/* revise le 06/10/2007 */
void SCE_CUseFramebuffer (SCE_CFramebuffer *fb, SCE_CIntRect *r)
{
    static int viewport[4];     /* viewport avant l'utilisation du fbo */
    static int binded_ = 0;     /* booleen, true: un fbo est deja binde */

    if (fb)
    {
        /* on recupere d'abord le viewport, pour ensuite le restituer
           (uniquement si aucun fbo n'est deja binde,
            sinon on prendrait son viewport) */
        if (!binded_)
            glGetIntegerv (GL_VIEWPORT, viewport);
        
        if (r != NULL)
            glViewport (r->x, r->y, r->w, r->h);
        else
            glViewport (fb->x, fb->y, fb->w, fb->h);
        
        glBindFramebufferEXT (GL_FRAMEBUFFER_EXT, fb->id);
        glClear (fb->mask);
        
        binded_ = SCE_TRUE;
    }
    else if (binded_)
    {
        glBindFramebufferEXT (GL_FRAMEBUFFER_EXT, 0);
        glViewport (viewport[0], viewport[1], viewport[2], viewport[3]);
        binded_ = SCE_FALSE;
    }
}
